#!/bin/bash
# ======================================================================
#
# IML BACKUP STORAGE HELPER
#
# ----------------------------------------------------------------------
#
# SYNTAX:
#   register [hostname]
#     init backup for hostname; backup client must interrupt/ repeat
#     return code is <> 0
#
#   unregister [hostname]
#     send info that a backup was done to delete the  connection flag
#     for the given host
#
#   status
#     show current connections
#
#   usage
#     show diskusage infos
#
# ----------------------------------------------------------------------
# 2019-11-01  hahn  write more clear log messages
# 2021-06-xx  hahn  added setactive + setinactive
# 2021-06-16  hahn  added backupstatus
# ======================================================================

# ----------------------------------------------------------------------
# CONFIG
# ----------------------------------------------------------------------


    . `dirname $0`/inc_config.sh

    # set in setBackupclient
    sBackupClient=dummy

    # value is set in _showConnectionCount
    iConnections=0

    # max age of last backup in h
    iMaxbackupAge=36

# ----------------------------------------------------------------------
# FUNCTIONS
# ----------------------------------------------------------------------


  # ------------------------------------------------------------
  # color text
  # ------------------------------------------------------------
    function color(){
      sColorcode=""
      case $1 in
        "reset") sColorcode="0"
                 ;;
        "head")  sColorcode="33" # yellow
                 ;;
        "cmd")   sColorcode="94" # light blue
                 ;;
        "input") sColorcode="92" # green
                 ;;
        "ok") sColorcode="92" # green
                 ;;
        "error") sColorcode="91" # red
                 ;;
      esac
      if [ ! -z ${sColorcode} ]; then
        echo -ne "\e[${sColorcode}m"
      fi
    }

    function _checkDir(){
      typeset -i iCount

      mydir=$1

      if [ -d $mydir ]; then
        iCount=`find $mydir -type f -cmin -720 | wc -l `
        if [ $iCount -eq 0 ]; then
          color error
        fi
        echo -n "$iCount        "
        echo -n "`du -hs $mydir #| awk ' { print $1 } '`        "

        if [ $iCount -eq 0 ]; then
          echo -n "<<< WARNING: No changed files detected !!! rc=1"
          color reset
          iErrors=$iErrors+1
        fi
        echo

      else
        echo SKIP: $mydir is not a directory
      fi
    }

  # ------------------------------------------------------------
  # logging
  # ------------------------------------------------------------
    # add a line into logfile
    # param  string  message (date will be added as prefix)
    function _addLog(){
        # echo `date` $* >>${sLogDir}${sLogfile}
        echo `date "+%Y-%m-%d %H:%M:%S"` $* >>${sLogfile}
    }

    # get content of all logfiles
    function _getAllLogs(){
        zcat $sLogdir/*.gz ; cat $sLogfile  
    }

    # get a list of all servers in the log
    function _getLoggedServers(){
        _getAllLogs | cut -f 2 -d '[' | cut -f 1 -d ']' | sort -u
    }

    # get timestamp of last backup start of a given server - see _getLoggedServers
    # param  string  servername as fqdn
    function _getLastBackupstart(){
        local _srv=$1
        _getAllLogs | sort | grep "\[$_srv\]" | grep ACCEPTED | tail -1 | cut -f -2 -d " "
    }
    
    # get timestamp of last backup end of a given server - see _getLoggedServers
    # param  string  servername as fqdn
    function _getLastBackupend(){
        local _srv=$1
        _getAllLogs | sort | grep "\[$_srv\]" | grep REMOVED | tail -1 | cut -f -2 -d " "
    }
  # ------------------------------------------------------------
  # slot handling
  # ------------------------------------------------------------

    # show current connction count
    function _showConnectionCount(){
        iConnections=`getConnectionCount`
        sMax=$iMaxConnections
        if [ $iMaxConnections -eq 0 ]; then
          sMax="0 [unlimited]"
        fi
        echo STATUS: $iConnections existing connection\(s\). Allowed maximum is $sMax.
    }

    
    # helper: get filename for backup client connection
    function getConnFilename(){
        # sBackupClient=$1
        echo $sConncectionDir/$sBackupClient
    }

    # allow a slot for a new connection
    # see PUBLIC_register
    function addConnection(){
        # sBackupClient=$1
        sTouchfile=`getConnFilename $sBackupClient`
        if [ -f "$sTouchfile" ]; then
            echo INFO: server [$sBackupClient] has a slot already
            _addLog "IGNORE request for a slot - [$sBackupClient] has a slot already"
        else
            touch "$sTouchfile"
            _addLog "REQUEST a slot for server [$sBackupClient]"
        fi
    }

    # remove slot of a connection
    # see PUBLIC_unregister
    function freeConnection(){
       #  sBackupClient=$1
        sTouchfile=`getConnFilename $sBackupClient`
        if [ -f "$sTouchfile" ]; then
            rm -f "$sTouchfile"
        else
            echo INFO: server [$sBackupClient] had no reserved slot
            _addLog "IGNORE freeing the connection - [$sBackupClient] had no reserved slot"
        fi
    }


    # count of conection slots
    function getConnectionCount(){
        # who | grep "^${sBackupUser}" | wc -l
        find $sConncectionDir -type f | wc -l
    }

    # helper function: set internal var for hostname of backup client
    function setBackupclient(){
        sBackupClient=`echo $1 | sed s"#[^a-zA-Z\.0-9]#_#g"`
        if [ -z $sBackupClient ]; then
            echo ERROR: no backup client was given
            exit 1
        fi
    }

    function _listRepodirs(){
        for mydir in $( ls -1d ${sBackupBasedir}/*/* | grep -vE '(_active|somethinglsetodisable)' )
        do
                echo -n "$mydir "
                test -f "$mydir/inactive.txt" && ( color error ; echo -n "(inactive)" )
                test -f "$mydir/inactive.txt" || ( color ok    ; echo -n "(active)"   )
                color reset
                echo
        done
    }

# ----------------------------------------------------------------------
# PUBLIC FUNCTIONS
# ----------------------------------------------------------------------


    function PUBLIC_backupstatus(){
        typeset -i local ierrors=0
        local tbl="%-45s | %-20s | %-20s | %-12s | %-10s \n"
        printf "$tbl" "server" "start" "end" "duration [s]" "age [h]"
        echo "--------------------------------------------------------------------------------------------------------------"
        for myserver in $( _getLoggedServers )
        do
            # get data
            local tstart="$( _getLastBackupstart $myserver )"
            local tend="$( _getLastBackupend $myserver )"
            typeset -i local istart=$( date +%s -d "$tstart" )
            typeset -i local iend=$( date +%s -d "$tend" )
            typeset -i local iduration=$iend-$istart
            typeset -i local iage=$( date +%s )-$iend
            typeset -i local iagehours=$iage/60/60
            
            # check values
            local statusAge=$(color ok)
            if [ $iagehours -gt $iMaxbackupAge ]; then
                statusAge=$(color error)
                ierrors=$ierrors+1
            fi
            
            local sduration=$iduration
            test $iduration -lt 0 && sduration="RUNNING"

            # output
            printf "$tbl" $myserver "$tstart" "${tend}" "${sduration}" "${statusAge}${iagehours}$(color reset)"
        done
        echo
        echo "total : $( _getLoggedServers | wc -l ) servers"
        echo "errors: $ierrors"
        echo
        echo "rc=$ierrors"
        exit $ierrors
    }

    function PUBLIC_setactive(){
        echo --- list of inactive backup repositories
        _listRepodirs | grep '\(inactive\)'
        echo
        echo -n "Repo to activate again >"
        read inputdir
        if [ -f "$inputdir/inactive.txt" ]; then
                cat "$inputdir/inactive.txt"
                echo
                echo Removing $inputdir/inactive.txt
                rm -f "$inputdir/inactive.txt"
                echo If it was a mistake run $( basename $0 ) setinactive 
        else
                echo SKIP [$inputdir]
        fi
        echo
    }

    function PUBLIC_setinactive(){
        echo --- list of backup repositories
        _listRepodirs
        echo
        echo -n "Repo to deactivate >"
        read inputdir
        if [ -d "$inputdir" -a ! -f "$inputdir/inactive.txt" ]; then
                echo -n "Short message >"
                read mymessage
                echo "$( date ) | $mymessage" >"$inputdir/inactive.txt"
                ls -l "$inputdir/inactive.txt"
        else
                echo SKIP [$inputdir]
        fi
        echo
    }

    # return backup directory for a client
    function PUBLIC_getBackupdir(){
        setBackupclient $1
        echo $sBackupBasedir/$sBackupClient
        exit 0
    }

    # register a hostname to start backups
    function PUBLIC_register(){
        setBackupclient $1
        echo "REGISTER $1"
        echo

        addConnection
        # iConnections=`getConnectionCount`
        _showConnectionCount >/dev/null

        if [ $iConnections -le $iMaxConnections -o $iMaxConnections -eq 0 ]; then
            sStatus="OK"
            iExit=0
            _addLog "ACCEPTED [$sBackupClient] got a slot."
        else
            sStatus="FAILED"
            iExit=1

            freeConnection
            _addLog "REJECTED maximum of $iMaxConnections connections exceeded - request of server [$sBackupClient] failed."
        fi

        _showConnectionCount
        echo $sStatus - rc=$iExit
        exit $iExit

    }

    # unregister a backup slot
    function PUBLIC_unregister(){
        setBackupclient $1
        echo "UNREGISTER $1"
        echo
        freeConnection
        _addLog "REMOVED slot for [$sBackupClient]"
        _showConnectionCount
        exit 0
    }

    # show used slots
    function PUBLIC_status(){
        _showConnectionCount
        echo
        if [ $iConnections -gt 0 ]; then
            find $sConncectionDir -type f -exec ls -l {} \; | nl
            echo
        fi
        exit 0
    }

    function PUBLIC_usage(){
        typeset -i iErrors=0

        # can be "full"
        sDetail=$1

        echo "DISK USAGE:"
        echo
        echo "time                   : `date`"
        echo "Backups are located in : $sBackupBasedir"
        echo "dirs                   : `find ${sBackupBasedir} -type d | wc -l`"
        echo "files                  : `find ${sBackupBasedir} -type f | wc -l`"
        echo "volume size            : `df -h ${sBackupBasedir} | tail -1 | awk '{ print $2 }'`"
        echo "backup size            : `du -hs ${sBackupBasedir} | cut -f 1 `"
        echo "free                   : `df -h ${sBackupBasedir} | tail -1 | awk '{ print $4 }'`"
        echo
        echo sizes:
        # ls -l "${sBackupBasedir}" | grep "^d" | fgrep -v "`basename $sConncectionDir`"
        du -hs ${sBackupBasedir}*
        echo

        echo "changed files last 12 hours and used diskspace:"
        echo
        _checkDir ${sBackupBasedir}
        echo
        if [ ! -z ${sDetail} ]; then
          sFilter=${sDetail}
          if [ "${sDetail}" = "full" ]; then
            sFilter=
          fi

          echo "... and by server and backup sets:"
          echo

          for myserver in `ls -1 ${sBackupBasedir} | fgrep -v "_active" | grep "$sFilter"`
          do
            color head
            echo ----- $myserver
            color reset
            echo
            _checkDir ${sBackupBasedir}/$myserver
            echo
            for subdir in `ls -1 ${sBackupBasedir}/$myserver`
            do
              _checkDir ${sBackupBasedir}/$myserver/$subdir
            done
            echo
          done
        fi

        # --- show result
        echo -n "Status: "
        if [ $iErrors -eq 0 ]; then
          echo "OK"
        else
          color error
          echo "$iErrors Errors detected."
        fi
        color reset
        echo rc=$iErrors
        exit $iErrors
    }

# ----------------------------------------------------------------------
# MAIN
# ----------------------------------------------------------------------


    if [ ! -d "$sConncectionDir" ]; then
        mkdir -p "$sConncectionDir" || exit 1
    fi

    color head
    echo
    echo "---------- :: STORAGE :: `hostname` :: ----------"
    echo
    color reset

    sFunction="PUBLIC_$1"
    fgrep "${sFunction}()" $0 >/dev/null


    if [ $? -ne 0 ]; then
        echo SYNTAX:
        echo "`basename $0` [function]"
        echo
        echo "  backupstatus"
        echo "    show all servers and their backup times"
        echo
        echo "  register | unregister [hostname]"
        echo "    add/ remove slot for a backup client"
        echo
        echo "  setactive"
        echo "    Reactivate an inctave backup repo."
        echo
        echo "  setinactive"
        echo "    Mark a backup repo as inactive."
        echo "    This prevents errors of missing backup data"
        echo "    if a host is obsolete and was shut down."
        echo
        echo "  status"
        echo "    show current reserved slots"
        echo
        echo "  usage [full|[filter to hostname]]"
        echo "    show used diskspace"
        echo "    you can set a string to show additionally details if these hosts"
        echo "    if you add \"full\" then details of all hosts will be shown"
        echo
        #echo SCAN... all functions are:
        #grep "function\ PUBLIC_" $0 | cut -f 2 -d "_" | cut -f 1 -d "("
    else
        shift 1
        $sFunction $*
    fi

# ----------------------------------------------------------------------
